(function()
{
	"use strict";
	
	var E_HORI = 0;
	var E_VERT = 1;
	var E_DIAG = 2;
	window.edgeTypes = {
		'E_HORI': E_HORI,
		'E_VERT': E_VERT,
		'E_DIAG': E_DIAG
		};
	
	function Edge(x1, y1, x2, y2, type)
	{
		this.x1 = x1;
		this.y1 = y1;
		this.x2 = x2;
		this.y2 = y2;
		this.type = type;
	}
	
	function boundingCoords(discs)
	{
		// compute minimum and maximum x and y needed to contain
		// all the discrepancies
		
		var w = discs.length;
		var h = discs[0].length;
		var inf = Number.POSITIVE_INFINITY;
		var minx = inf;
		var miny = inf;
		var maxx = -inf;
		var maxy = -inf;
		for(var i = 0; i < w; i++)
		{
			for(var j = 0; j < h; j++)
			{
				if(discs[i][j] != 0)
				{
					if(i < minx) minx = i;
					if(i > maxx) maxx = i;
					if(j < miny) miny = j;
					if(j > maxy) maxy = j;
				}
			}
		}
		return [minx, miny, maxx, maxy];
	}
	window.boundingCoords = boundingCoords;
	
	function getEdgeSet(discs)
	{
		// compute the set of edges from an array of discrepancies
		var bc = boundingCoords(discs);
		
		var edgeList = [];
		
		for(var i = bc[0]; i <= bc[2] + 1; i++)
		{
			for(var j = bc[1]; j <= bc[3] + 1; j++)
			{
				// add all bounding edge
				if(i <= bc[2]) edgeList.push(new Edge(i, j, i + 1, j, E_HORI));
				if(j <= bc[3]) edgeList.push(new Edge(i, j, i, j + 1, E_VERT));
				if(i <= bc[2] && j <= bc[3] && discs[i][j] == 1) // minor discrepancy
				{
					// also add diagonals
					edgeList.push(new Edge(i, j, i + 1, j + 1, E_DIAG));
					edgeList.push(new Edge(i + 1, j, i, j + 1, E_DIAG));
				}
			}
		}
		// filter unique
		var uniqueDict = {};
		var s;
		for(var k = 0; k < edgeList.length; k++)
		{
			s = JSON.stringify(edgeList[k]);
			uniqueDict[s] = edgeList[k];
		}
		edgeList = []
		for(s in uniqueDict)
		{
			if(({}).hasOwnProperty.call(uniqueDict, s))
			{
				edgeList.push(uniqueDict[s]);
			}
		}
		return edgeList;
	}
	window.getEdgeSet = getEdgeSet;
	
	function distLineSegToPoint(x1, y1, x2, y2, x, y)
	{
		var pdispx = x - x1;
		var pdispy = y - y1;
		var lsdispx = x2 - x1;
		var lsdispy = y2 - y1;
		var projFactor = pdispx * lsdispx + pdispy * lsdispy;
		var normsqls = (lsdispx * lsdispx + lsdispy * lsdispy);
		projFactor /= normsqls;
		var projx = lsdispx * projFactor;
		var projy = lsdispy * projFactor;
		var orthx = pdispx - projx;
		var orthy = pdispy - projy;
		var result;
		if(projFactor > 0 && projFactor < 1)
		{
			// distance to line seg = distance to line
			result = Math.sqrt(orthx * orthx + orthy * orthy);
		}
		else
		{
			// distance to line seg = shortest distance to point
			var d1sq = pdispx * pdispx + pdispy * pdispy;
			var pdispx2 = x - x2;
			var pdispy2 = y - y2;
			var d2sq = pdispx2 * pdispx2 + pdispy2 * pdispy2;
			result = Math.sqrt(Math.min(d1sq, d2sq));
		}
		return result;
	}
	
	function getNearbyEdge(edgeSet, x, y)
	{
		var cutoff = 0.25;
		// find a suitable edge for edge picking
		var mind = Number.POSITIVE_INFINITY;
		var mine;
		for(var n = 0; n < edgeSet.length; n++)
		{
			var e = edgeSet[n];
			var d = distLineSegToPoint(e.x1, e.y1, e.x2, e.y2, x, y);
			if(d < mind)
			{
				mind = d;
				mine = e;
			}
		}
		if(mind < cutoff)
		{
			return mine;
		}
		else
		{
			return null;
		}
	}
	window.getNearbyEdge = getNearbyEdge;
	
	function diagonallyCrossingEdgeArray(edgeSet, w, h)
	{
		// produce a 2d array counting the number of selected
		// diagonal edges in each cell
		var result = util.grid2d(w, h, 0);
		for(var eIdx = 0; eIdx < edgeSet.length; eIdx++)
		{
			var e = edgeSet[eIdx];
			if(e.type != E_DIAG) continue;
			var cellx = Math.min(e.x1, e.x2);
			var celly = Math.min(e.y1, e.y2);
			result[cellx][celly]++;
		}
		return result;
	}
	
	function countLoops(edgeSet, diagCount)
	{
		// test that the edge set forms a set of loops
		// and if so, return the number of loops. if not
		// return null
		
		// count the valency of every vertex - should be zero or 2
		var vertDeg = {};
		var eIdx;
		for(eIdx = 0; eIdx < edgeSet.length; eIdx++)
		{
			var e = edgeSet[eIdx];
			var key1 = JSON.stringify([e.x1, e.y1]);
			var key2 = JSON.stringify([e.x2, e.y2]);
			if(!vertDeg.hasOwnProperty(key1)) vertDeg[key1] = 0;
			vertDeg[key1]++;
			if(!vertDeg.hasOwnProperty(key2)) vertDeg[key2] = 0;
			vertDeg[key2]++;
		}
		for(var vKey in vertDeg)
		{
			if(vertDeg.hasOwnProperty(vKey))
			{
				if(vertDeg[vKey] != 2)
				{
					// vertex with valency other than 2 - failed
					return null;
				}
			}
		}
		
		// check that the two diagonals of any cell are not both included
		for(var i = 0; i < diagCount.length; i++)
		{
			for(var j = 0; j < diagCount[i].length; j++)
			{
				if(diagCount[i][j] > 1)
				{
					// two crossing diagonals selected - failed
					return null;
				}
			}
		}
		
		// use a region-growing method to count the number of loops
		// this is NOT EFFICIENT. An EFFICIENT solution would work by
		// keeping track of islands and merging them but now isn't the time :)
		
		// set up a list of edges remaining
		var edgesRemaining = [];
		for(eIdx = 0; eIdx < edgeSet.length; eIdx++)
		{
			edgesRemaining.push(edgeSet[eIdx]);
		}
		
		// while any edges still remain...
		var numLoops = 0;
		var eIdx1, eIdx2;
		while(edgesRemaining.length > 0)
		{
			// extract one
			var curLoop = edgesRemaining.splice(0, 1);
			numLoops++;
			// while not done...
			var done = false
			while(!done)
			{
				// look for an edge sharing a vertex with one in the current set
				var foundIdx = -1;
				for(eIdx1 = 0; eIdx1 < curLoop.length; eIdx1++)
				{
					var e1 = curLoop[eIdx1];
					for(eIdx2 = 0; eIdx2 < edgesRemaining.length; eIdx2++)
					{
						var e2 = edgesRemaining[eIdx2];
						if(e1.x1 == e2.x1 && e1.y1 == e2.y1 ||
							e1.x2 == e2.x2 && e1.y2 == e2.y2 ||
							e1.x1 == e2.x2 && e1.y1 == e2.y2 ||
							e1.x2 == e2.x1 && e1.y2 == e2.y1)
						{
							foundIdx = eIdx2;
							break;
						}
					}
					if(foundIdx != -1) break;
				}
				// if found, extract from remaining edges and add to current set
				if(foundIdx != -1)
				{
					curLoop.push(edgesRemaining.splice(foundIdx, 1)[0]);
				}
				else // if not found, we are done.
				{
					done = true;
				}
			}
		}
		// return the number of loops found
		return numLoops;
	}
	window.countLoops = countLoops;
	
	function cellEntirelyContained(edges, x, y)
	{
		// assuming that a set of edges forms a set of loops,
		// this function determines whether a cell is wholly
		// contained in a loop.
		
		// idea: find how many edges of the loops lie "above" the midpoint of the cell,
		// check if it's odd
		var count = 0
		for(var eIdx = 0; eIdx < edges.length; eIdx++)
		{
			var e = edges[eIdx];
			if(Math.min(e.x1, e.x2) <= x && Math.max(e.x1, e.x2) > x &&
				Math.max(e.y1, e.y2) <= y)
			{
				count++;
			}
		}
		return ((count % 2) != 0);
	}
	
	function isLevelComplete(edges, disc, goalNumLoops)
	{
		return isLevelCompleteAndWhyNot(edges, disc, goalNumLoops)[0];
	}
	
	window.isLevelCompleteAndWhyNot = function(edges, disc, goalNumLoops)
	{
		var w = disc.length;
		var h = disc[0].length;
		var diagCount = diagonallyCrossingEdgeArray(edges, w, h);
		var nLoops = countLoops(edges, diagCount);
		if(nLoops === null || nLoops === 0)
		{
			// the player hasn't drawn any loops yet. don't provide a message
			return [false, null]
		}
		
		// check if major discrepancies are contained, and normal
		// memory is not
		var falsePos = 0;
		var falseNeg = 0;
		var i, j;
		for(i = 0; i < w; i++)
		{
			for(j = 0; j < h; j++)
			{
				if(disc[i][j] == 1) continue;
				var isContained = cellEntirelyContained(edges, i, j);
				if(disc[i][j] == 0 && isContained) falsePos++;
				if(disc[i][j] == 2 && !isContained) falseNeg++;
			}
		}
		
		// check if minor discrepancies are cut in two
		var nMinorsNotCut = 0;
		for(i = 0; i < w; i++)
		{
			for(j = 0; j < h; j++)
			{
				if(disc[i][j] == 1)
				{
					if(diagCount[i][j] != 1)
					{
						nMinorsNotCut++;
					}
				}
			}
		}
		
		var fail1 = (falseNeg != 0);
		var fail2 = (nMinorsNotCut != 0);
		var fail3 = (falsePos != 0);
		var fail4 = (nLoops != goalNumLoops);
		
		var hint = null;
		var failCount = fail1 + fail2 + fail3 + fail4;
		if (failCount == 1)
		{
			// only provide a hint if only one thing has gone wrong
			if(fail1) hint = "Not all major discrepancies are isolated.";
			if(fail2) hint = "Not all minor discrepancies are cut.";
			if(fail3) hint = "Valid memory should not be isolated.";
			if(fail4) hint = ("There should be " + goalNumLoops + 
				(goalNumLoops==1?" loop.":" loops."));
		}
		
		if(failCount != 0)
		{
			return [false, hint];
		}
		
		return [true, null];
	}
	window.isLevelComplete = isLevelComplete;
})();


