(function()
{
	"use strict";

	var levelGen = {};
	window.levelGen = levelGen;
	
	levelGen.w = 10;
	levelGen.h = 10;
	var GLOBAL_W = levelGen.w;
	var GLOBAL_H = levelGen.h;
	
	levelGen.allQuarters = function(w, h)
	{
		var result = [];
		for(var i = 0; i < w; i++)
		{
			for(var j = 0; j < h; j++)
			{
				for(var k = 0; k < 4; k++)
				{
					result.push(new Quarter(i, j, k));
				}
			}
		}
		return result;
	}
	
	levelGen.genQSet = function(nComponents, w, h, maxExpands)
	{
		var done = false;
		var maxTrials = 25;
		var minExpands = 5;
		var qSet;
		var i, j, r, q, m;
		var allQuarters = levelGen.allQuarters(w, h);
		if(w > GLOBAL_W || h > GLOBAL_H)
		{
			throw new Error("Input sizes are out of valid range");
		}
		
		while(!done)
		{
			qSet = new QuarterSet();
			
			// First, generate a qset with a few isolated halves
			
			var trialNum = 0;
			var ok = true;
			for(var comp = 0; comp < nComponents; comp++)
			{
				ok = false;
				while(!ok && trialNum < maxTrials)
				{
					i = Math.floor(Math.random() * w);
					j = Math.floor(Math.random() * h);
					r = Math.floor(Math.random() * 4);
					q = new Quarter(i, j, r);
					if(qSet.isHalfLonely(q))
					{
						qSet.grid[i][j][r] = true;
						qSet.grid[i][j][(r + 1) % 4] = true;
						ok = true;
					}
					trialNum++;
				}
				if(!ok) break;
			}
			if(!ok) continue;
			
			// now try to expand
			var expandNum = 0;
			
			while(true)
			{
				if(expandNum >= maxExpands) break;
				var eAbles = [];
				for(m = 0; m < allQuarters.length; m++)
				{
					if(qSet.canGrowToHalf(allQuarters[m]))
					{
						eAbles.push(allQuarters[m]);
					}
				}
				if(eAbles.length==0) break;
				m = Math.floor(Math.random() * eAbles.length);
				i = eAbles[m].x;
				j = eAbles[m].y;
				r = eAbles[m].r;
				qSet.grid[i][j][r] = true;
				qSet.grid[i][j][(r + 1) % 4] = true;
				expandNum++;
			}
			
			done = true;
		}
		return qSet
	}
	
	levelGen.qSetToDiscs = function(qSet)
	{
		// return an array filled with 0, 1, 2 for the type of discrepancy
		var g = qSet.grid;
		var result = util.grid2d(GLOBAL_W, GLOBAL_H)
		for(var i = 0; i < GLOBAL_W; i++)
		{
			for(var j = 0; j < GLOBAL_H; j++)
			{
				var nQs = 0;
				for(var k = 0; k < 4; k++)
				{
					nQs += g[i][j][k];
				}
				if((nQs % 2) == 0)
				{
					result[i][j] = nQs / 2;
				}
				else
				{
					throw new Error("Sanity check failed - filled space is not half-integer");
				}
			}
		}
		return result;
	}
	
	levelGen.genHardDiscs = function(nComponents, w, h, maxMaxExtends)
	{
		var bestDiscs, bestX;
		var bestOnes = -1;
		for(var n = 0; n < 50; n++)
		{
			var maxExtend = 5 + Math.floor(Math.random() * (maxMaxExtends - 5));
			var x = levelGen.genQSet(nComponents, w, h, maxExtend);
			var discs = levelGen.qSetToDiscs(x);
			var ones = 0;
			for(var i = 0; i < discs.length; i++)
			{
				for(var j = 0; j < discs[i].length; j++)
				{
					if(discs[i][j] == 1) ones++;
				}
			}
			if(ones > bestOnes)
			{
				bestOnes = ones;
				bestDiscs = discs;
				bestX = x;
			}
		}
		return bestDiscs;
	}
	
	function Quarter(x,y,r)
	{
		this.x = x;
		this.y = y;
		this.r = r;
	};
	window.Quarter = Quarter;
	
	Quarter.prototype.mate = function()
	{
		var newX = this.x;
		var newY = this.y;
		var r = this.r;
		var newR = (r + 2) % 4;
		if(r == 0) newY -= 1;
		if(r == 1) newX += 1;
		if(r == 2) newY += 1;
		if(r == 3) newX -= 1;
		return new Quarter(newX, newY, newR);
	}
	
	Quarter.prototype.next = function()
	{
		return new Quarter(this.x, this.y, (this.r + 1) % 4);
	}
	
	Quarter.prototype.prev = function()
	{
		return Quarter(x, y, (r + 3) % 4);
	}
	
	function inBounds(x, y)
	{
		if(x < 0 || x >= GLOBAL_W) return false;
		if(y < 0 || y >= GLOBAL_H) return false;
		return true;
	}
	
	function QuarterSet()
	{
		this.grid = util.grid3d(levelGen.w, levelGen.h, 4, false);
	}
	window.QuarterSet = QuarterSet;
	
	function occupiedCoords(grid, x, y, r)
	{
		if(!inBounds(x, y)) return false;
		return grid[x][y][r];
	}
	
	function occupiedQuarter(grid, quarter)
	{
		var x = quarter.x;
		var y = quarter.y;
		var r = quarter.r;
		return occupiedCoords(grid, x, y, r);
	}
	
	QuarterSet.prototype.isCornerOccupiedByQuarter = function(quarter)
	{
		var g = this.grid;
		for(var i = 0; i < 4; i++)
		{
			if(occupiedQuarter(g, quarter)) {return true;}
			quarter = quarter.mate();
			if(occupiedQuarter(g, quarter)) {return true;}
			quarter = quarter.next();
		}
		return false;
	}
	
	QuarterSet.prototype.isHalfLonely = function(firstQuarter)
	{
		// Box-halves can be indexed by the same data as box-quarters.
		// So we represent a box-half by its first box-quarter.
		// Determine whether this half is isolated from all occupied quarters
		if(this.isCornerOccupiedByQuarter(firstQuarter)) return false;
		firstQuarter = firstQuarter.next();
		if(this.isCornerOccupiedByQuarter(firstQuarter)) return false;
		firstQuarter = firstQuarter.next();
		if(this.isCornerOccupiedByQuarter(firstQuarter)) return false;
		return true;
	}
	
	QuarterSet.prototype.canGrowToHalf = function(firstQuarter)
	{
		// represent a box-half by its first box-quarter
		// determine whether the quarterset can be validly grown to have
		// this half. it should be empty, and EITHER:
		// - it shares an orthogonal edge with one occupied quarter, and
		//   the opposite corner is on no occupied quarters, OR
		// - it shares the diagonal edge with an occupied half, and the
		//   opposite corner is on no occupied corners, OR
		// - it shares two edges
		var g = this.grid;
		var q0 = firstQuarter;
		var q1 = q0.next();
		var q2 = q1.next();
		var q3 = q2.next();
		// check that this half is empty
		var selfOcc = (occupiedQuarter(g, q0) || occupiedQuarter(g, q1));
		if(selfOcc) return false;
		// determine if the neighbours along the diagonal are occupied
		var diagOcc = (occupiedQuarter(g, q2) && occupiedQuarter(g, q3));
		// determine if the neighbour along the orthogonal edges are occupied
		var mateOcc0 = occupiedQuarter(g, q0.mate());
		var mateOcc1 = occupiedQuarter(g, q1.mate());
		// if at least two of the neighbours are occupied, we're done
		var numNeighbs = diagOcc + mateOcc0 + mateOcc1;
		if(numNeighbs == 3)
		{
			throw new Error("Sanity check failed, empty triangle with 3 " +
				"occupied neighbours")
		}
		if(numNeighbs == 2) return true;
		if(numNeighbs == 0) return false;
		// there is one neighbour. determine if the corners are empty
		if(mateOcc0)
		{
			return !this.isCornerOccupiedByQuarter(q2);
		}
		if(mateOcc1)
		{
			return !this.isCornerOccupiedByQuarter(q0);
		}
		if(diagOcc)
		{
			return !this.isCornerOccupiedByQuarter(q1);
		}
		throw new Error("Sanity check failed, could not determine corner to test");
	}
	
})();

