(function(){
	"use strict";
	
	var MUSIC_VOLUME = 0.25;
	var SPEECH_VOLUME = 0.75;
	var SPEECH_VOLUME_REPEAT = 0.6;
	var PRE_MUSIC_DELAY = 3; // seconds
    
	function Resources()
	{
		// constructor for a collection of resources
		this.data = {};
		this.amtLoaded = {};
		this.requests = [];
		this.sizes = {};
		this.numSizesGot = 0;
		this.numRequests = 0;
		this.numDownloaded = 0;
		this.numDecoded = 0;
		this.totalLoaded = 0;
		this.totalToLoad = 0;
		this.playlist = [];
		this.currentTrackId = -1;
		this.musicStarted = false;
		this.speechSequence = null;
		this.speechIndex = null;
		this.speechTimeout = null;
		this.speechRepeatMode = false;
		this.lastSpeechEndedTime = 0;
		this.repeatCount = 0;
        
        window.AudioContext = window.AudioContext || window.webkitAudioContext;
        
        if(window.AudioContext)
        {
            this.useWebAudio = true;
            this.audioContext = new AudioContext();
            
            this.musicGainNode = this.audioContext.createGain();
            this.musicGainNode.gain.value = MUSIC_VOLUME;
            this.musicGainNode.connect(this.audioContext.destination);
            
            this.speechGainNode = this.audioContext.createGain();
            this.speechGainNode.gain.value = SPEECH_VOLUME;
            this.speechGainNode.connect(this.audioContext.destination);
            
            this.speechRepeatGainNode = this.audioContext.createGain();
            this.speechRepeatGainNode.gain.value = SPEECH_VOLUME_REPEAT;
            this.speechRepeatGainNode.connect(this.audioContext.destination);
        }
        else
        {
            this.useWebAudio = false;
        }
	}
	window.Resources = Resources;
	
	Resources.prototype.reqImage = function(name)
	{
		// request an image
		this.requests.push([name, ".png", "image/png"]);
		this.numRequests++;
	}
	
	Resources.prototype.reqAudio = function(name)
	{
		// request an mp3 sound
		this.requests.push([name, ".mp3", "audio/mpeg"]);
		this.numRequests++;
	}
	
	Resources.prototype.getAllFileSizes = function(callback)
	{
		// start getting all the file sizes
		for(var i = 0; i < this.requests.length; i++)
		{
			var name = this.requests[i][0];
			var ext = this.requests[i][1];
			var mime = this.requests[i][2];
			var filename = name + ext;
			this.getFileSize(filename, mime, 
				this.handleGotFileSize.bind(this, name, callback));
		}
	}
	
	Resources.prototype.getFileSize = function(filename, mime, callback)
	{
		var req = new XMLHttpRequest();
		req.open("HEAD", filename);
		req.overrideMimeType(mime);
		req.onreadystatechange = function() {
			if(req.readyState == req.DONE)
			{
				var bytes = parseInt(req.getResponseHeader("Content-Length"));
				callback(bytes);
			}
		}
		req.send();
	}
	
	Resources.prototype.handleGotFileSize = function(name, callback, size)
	{
		// handle receiving a file size
		this.sizes[name] = size;
		this.numSizesGot++;
		if(this.numSizesGot == this.numRequests)
		{
			this.recomputeLoaded();
			callback();
		}
	}
	
	Resources.prototype.downloadAll = function(callback)
	{
		// start getting all the file sizes
		for(var i = 0; i < this.requests.length; i++)
		{
			var name = this.requests[i][0];
			var ext = this.requests[i][1];
			var mime = this.requests[i][2];
			this.amtLoaded[name] = 0;
			var filename = name + ext;
			this.downloadFile(name, filename, mime,
				this.handleDownloadedFile.bind(this, name, mime, callback));
		}
	}
	
	Resources.prototype.recomputeLoaded = function()
	{
		var denom = 0;
		for(var x in this.sizes)
		{
			if(this.sizes.hasOwnProperty(x))
			{
				denom += this.sizes[x];
			}
		}
		this.totalToLoad = denom;
		
		var numer = 0;
		for(var x in this.amtLoaded)
		{
			if(this.amtLoaded.hasOwnProperty(x))
			{
				numer += this.amtLoaded[x];
			}
		}
		this.totalLoaded = numer;
	}
	
	Resources.prototype.downloadFile = function(name, filename, mime, callback)
	{
		var req = new XMLHttpRequest();
		req.open("GET", filename);
		req.overrideMimeType(mime);
		req.responseType = 'arraybuffer';
		req.onprogress = (function(evt) {
			this.amtLoaded[name] = evt.loaded;
			this.recomputeLoaded();
		}).bind(this);
		req.onreadystatechange = function() {
			if(req.readyState == req.DONE)
			{
				callback(req.response);
			}
		}
		req.send();
	}
	
	Resources.prototype.handleDownloadedFile = function(name, mime, callback, response)
	{
		this.numDownloaded++;
		if(mime == "image/png")
		{
			var blob = new Blob([response], {type: mime});
			var result = new Image();
			result.src = URL.createObjectURL(blob);
			result.onload = this.handleProcessedFile.bind(this, callback);
			this.data[name] = result;
		}
		else if(mime == "audio/mpeg")
		{
			if(this.useWebAudio)
			{
				// web audio API
				this.audioContext.decodeAudioData(response, (function(buffer)
				{
					this.data[name] = buffer;
					this.handleProcessedFile(callback);
				}).bind(this), function() {throw new Error('Could not decode sound');});
			}
			else
			{
				// regular audio
				var blob = new Blob([response], {type: mime});
				var result = new Audio();
				result.src = URL.createObjectURL(blob);
				result.addEventListener('canplaythrough',
					this.handleProcessedFile.bind(this, callback));
				this.data[name] = result;
			}
		}
		else
		{
			throw new Error("Don't know how to handle " + mime);
		}
	}
	
	Resources.prototype.handleProcessedFile = function(callback)
	{
		this.numDecoded++;
		if(this.numDecoded == this.numRequests)
		{
			callback();
		}
	}
	
	Resources.prototype.startMusic = function(playlist, firstTrackId)
	{
		this.musicStarted = true;
		this.playlist = playlist;
		
		this.currentTrackId = firstTrackId;
		var track = playlist[firstTrackId];
		
		if(this.useWebAudio)
		{
			var source = this.audioContext.createBufferSource();
			source.buffer = track;
			source.loop = false;
			source.connect(this.musicGainNode);
			source.start(this.audioContext.currentTime + PRE_MUSIC_DELAY);
			source.onended = this.nextTrack.bind(this);
		}
		else
		{
			track.currentTime = 0;
			track.volume = 0;
			setTimeout(function()
			{
				track.volume = MUSIC_VOLUME;
				track.play();
			}, PRE_MUSIC_DELAY * 1000);
			
			// give all the tracks event listeners to make them play the next track
			for(var i = 0; i < playlist.length; i++)
			{
				playlist[i].addEventListener('ended', this.nextTrack.bind(this), false);
			}
		}
	}
	
	Resources.prototype.nextTrack = function()
	{
		// select a random track other than the current one
		var nextTrackId = Math.floor(Math.random() * (this.playlist.length - 1));
		if(nextTrackId == this.currentTrackId) nextTrackId++;
		var track = this.playlist[nextTrackId];
		this.currentTrackId = nextTrackId;
		
		if(this.useWebAudio)
		{
			var source = this.audioContext.createBufferSource();
			source.buffer = track;
			source.loop = false;
			source.connect(this.musicGainNode);
			source.start(this.audioContext.currentTime + PRE_MUSIC_DELAY);
			source.onended = this.nextTrack.bind(this);
		}
		else
		{
			track.currentTime = 0;
			track.volume = 0;
			setTimeout(function()
			{
				track.volume = MUSIC_VOLUME;
				track.play();
			}, 3000);
		}
	}
	
	Resources.prototype.continueSpeech = function()
	{
		// check if there is any speech left for this level and if so,
		// say it
		
		if(this.speechTimeout !== null)
		{
			this.speechTimeout = null;
			this.lastSpeechEndedTime = timeSecs();
		}
		this.speechIndex++;
		var ss = this.speechSequence;
		var si = this.speechIndex;
		var shouldRepeat = (this.speechRepeatMode & this.repeatCount > 1);
		if(ss === null || si === null || (si >= ss.length && !shouldRepeat))
		{
			this.speechSequence = null;
			this.si = null;
			return;
		}
		var isRepeat = false;
		if(si >= ss.length) // repeatmode
		{
			si = 0;
			this.speechIndex = 0;
			isRepeat = true;
			this.repeatCount--;
		}
		var sound = this.data[ss[si][0]];
		this.speechText = ss[si][1];
        
        this.speechTimeout = setTimeout(this.continueSpeech.bind(this), 6000);
	}
	
	Resources.prototype.playSpeechSequence = function(sequence)
	{
		// play a sequence of speech
		this.speechRepeatMode = false;
		this.speechSequence = sequence;
		this.speechIndex = -1;
		this.continueSpeech();
	}
	
	Resources.prototype.playRepeatingSpeechSequence = function(sequence)
	{
		// play a repeating sequence of speech until the level is over
		this.speechRepeatMode = true;
		this.repeatCount = 3;
		this.speechSequence = sequence;
		this.speechIndex = -1;
		this.continueSpeech();
	}
	
	
	Resources.prototype.haltSpeechSequence = function(sequence)
	{
		this.speechRepeatMode = false;
		this.speechSequence = null;
		this.speechIndex = null;
		if(this.speechTimeout !== null)
		{
			clearTimeout(this.speechTimeout);
            this.speechTimeout = null;
		}
	}
    
    
	Resources.prototype.playSound = function(sound, callback)
	{
		if(this.useWebAudio)
		{
			var source = this.audioContext.createBufferSource();
			source.buffer = sound;
			source.loop = false;
            source.connect(this.speechGainNode);
			source.start(this.audioContext.currentTime);
			// call me again when sound finished
			this.speechTimeout = setTimeout(callback, (0.1  + sound.duration) * 1000);
			this.playingSpeech = source;
		}
		else
		{
			sound.currentTime = 0;
            sound.volume = SPEECH_VOLUME;
			sound.play();
			sound.onended = callback;
			this.playingSpeech = sound;
		}
	}
	
})();

